/*
SHG -- Spherical Harmonics Generator.
Copyright (C) 2005, 2006 Adomas Paltanavicius

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

// LGPL-ed zoom, pan and rotate from http://www.nigels.com/glt/gltzpr/
// by Nigel Stewart
#include "zpr.h"

// parameters
int argL, argM;

// vizualization parameters
int showLines;
int showOdd;

// computed data
int gridSize = 200;
double **grid;
char **odd;
int _gridSize;

// movement
GLdouble rotatex;
GLdouble rotatey;
GLdouble rotatez;

extern void Yslice(unsigned int, unsigned int, double, int, double *, char *);

void displayList();

void
compute()
{
	int i;
	glutSetCursor(GLUT_CURSOR_WAIT);
	printf("Computing grid[l=%d,m=%d,size=%d]... ", argL, argM, gridSize);
	// free grid, because gridSize might have changed
	if (grid) {
		for (i = 0; i < _gridSize; i++) {
			free(grid[i]);
			free(odd[i]);
		}
		free(grid);
		free(odd);
	}
	_gridSize = gridSize; // save for freeing
	grid = malloc(sizeof(grid[0]) * gridSize);
	odd = malloc(sizeof(odd[0]) * gridSize);
	for (i = 0; i < gridSize; i++) {
		double theta = (double)i/(double)gridSize * M_PI * 2.0;
		grid[i] = malloc(sizeof(grid[0][0]) * gridSize);
		odd[i] = malloc(sizeof(odd[0][0]) * gridSize);
		Yslice(argL, argM, theta, gridSize, &grid[i][0], &odd[i][0]);
		printf("+");
	}
	printf(" done.\n");
	glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
	displayList();
}

// display list
GLuint orbitListId;

GLdouble color1[3] = {0.7,0.3,0.3};
GLdouble color2[3] = {0.3,0.7,0.3};

void 
displayList() 
{
	int i, j;

	printf("Creating display list... ");

	if (orbitListId) {
		glDeleteLists(orbitListId, 1);
	}

	orbitListId = glGenLists(1);
	glNewList(orbitListId, GL_COMPILE);

	glPushMatrix();
	for (i = 0; i < gridSize; i++) { // vert
		double theta1 = (double)i/(double)gridSize * M_PI * 2.0;
		double theta2 = (double)((i+1)%gridSize)/(double)gridSize * M_PI * 2.0;
		if (showLines) {
			glBegin(GL_LINES);
		} else {
			glBegin(GL_QUADS);
		}
		for (j = 0; j < gridSize; j++) { // horz
			double phi1 = (double)j/(double)gridSize * M_PI * 2.0;
			double phi2 = (double)((j+1)%gridSize)/(double)gridSize * M_PI * 2.0;
			double rho1 = grid[i][j];
			double rho2 = grid[i][(j+1)%gridSize];
			double rho3 = grid[(i+1)%gridSize][j];
			double rho4 = grid[(i+1)%gridSize][(j+1)%gridSize];
			GLdouble n[3];

			GLdouble p1[3] = {rho1 * sin(theta1) * cos(phi1), rho1 * sin(theta1) * sin(phi1), rho1 * cos(theta1)};
			GLdouble p2[3] = {rho2 * sin(theta1) * cos(phi2), rho2 * sin(theta1) * sin(phi2), rho2 * cos(theta1)};
			GLdouble p3[3] = {rho4 * sin(theta2) * cos(phi2), rho4 * sin(theta2) * sin(phi2), rho4 * cos(theta2)};
			GLdouble p4[3] = {rho3 * sin(theta2) * cos(phi1), rho3 * sin(theta2) * sin(phi1), rho3 * cos(theta2)};

			// one quad
		    if (showOdd && odd[i][j]) {
				glColor3dv(color1);
			} else {
				glColor3dv(color2);
			}
			// normal
			// skaiciuojam kaip (0,0,0) -> (p1+p2+p3+p4)/4
			n[0] = (p1[0]+p2[0]+p3[0]+p4[0])/4.;
			n[1] = (p1[1]+p2[1]+p3[1]+p4[1])/4.;
			n[2] = (p1[2]+p2[2]+p3[2]+p4[2])/4.;
			glNormal3dv(n);
			if (showLines) {
				glVertex3dv(p1);
				glVertex3dv(p3);
				glVertex3dv(p2);
				glVertex3dv(p4);
			} else {
				glVertex3dv(p1);
				glVertex3dv(p2);
				glVertex3dv(p3);
				glVertex3dv(p4);
			}
		}
		glEnd();
	}
	glPopMatrix();
	glEndList();

	printf("done\n");
}

#if 0
void normal() {

}
#endif

void
display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glRotated(rotatex, 1.0, 0.0, 0.0);
	glRotated(rotatey, 0.0, 1.0, 0.0);
	glRotated(rotatez, 0.0, 0.0, 1.0);
	glCallList(orbitListId);
    glutSwapBuffers();
}

void
init(void)
{
	GLfloat light_ambient[]  = { 0.0, 0.0, 0.0, 1.0 };
	GLfloat light_diffuse[]  = { 1.5, 1.5, 1.5, 1.5 };
	GLfloat light_specular[] = { .5, .5, .5, .5 };
	GLfloat light_position[] = { -3.0, 3.0, 3.0, 0.0 };

	GLfloat mat_ambient[]    = { 0.7, 0.7, 0.7, 1.0 };
	GLfloat mat_diffuse[]    = { 0.8, 0.8, 0.8, 1.0 };
	GLfloat mat_specular[]   = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat high_shininess[] = { 15.0 };

	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);

	glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    	glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    	glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_POLYGON_SMOOTH);
	glEnable(GL_DEPTH_TEST);
    	glEnable(GL_NORMALIZE);
    	glEnable(GL_COLOR_MATERIAL);
	glShadeModel(GL_SMOOTH);

	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

int doAnimate;

static void
idle(void)
{
	if (doAnimate) {
		rotatex = 0.3;
		rotatey = 0.3;
		glutPostRedisplay();
	}
}

void
key(unsigned char k, int x, int y)
{
   switch (k) {
   case 27:
      exit(0 + x-x + y-y); //hehe
      break;
   case 'f':
      glutFullScreen(); // cool - not only fullscreen, but also somehow doesn't show X button, so esc always exit(0)s
   }
   glutPostRedisplay();  
}

/* menu */
char *menuItems[] = {
	"Quit",
	"l+",
	"l-",
	"m+",
	"m-",
	"Toggle lines/surface",
	"Toggle multicolor (odd)",
	"Toggle animation",
	"grid+5",
	"grid+10",
	"grid+20",
	"grid+50",
	"grid-5",
	"grid-10",
	"grid-20",
	"grid-50",
};

int nMenuItems = sizeof(menuItems)/sizeof(menuItems[0]);
const int nMenuSubshell = 15;

void
menuData(int value) {
	if (value < nMenuItems) {
		switch (value) {
			case 0: exit(0);
			case 1: argL++; break;
			case 2: 
				if (argL) argL--;
				if (argM>argL || argM<-argL)
					argM = 0;
				break;
			case 3: if (argM<argL) argM++; break;
			case 4: if (argM>-argL) argM++; break;
			case 5: showLines = !showLines; break;
			case 6: showLines = 0; showOdd = !showOdd; break;
			case 7: doAnimate = !doAnimate; break;
			case 8: gridSize += 5; break;
			case 9: gridSize += 10; break;
			case 10: gridSize += 20; break;
			case 11: gridSize += 50; break;
			case 12: if (gridSize > 5) gridSize -= 5; break;
			case 13: if (gridSize > 10) gridSize -= 10; break;
			case 14: if (gridSize > 20) gridSize -= 20; break;
			case 15: if (gridSize > 50) gridSize -= 50; break;
		}
	} else {
		value -= nMenuItems;
		argL = argM = 0;
		while (value >= argL*2+1) {
			value -= argL*2+1;
			argL++;
		}
		argM = value - argL;
	}
	compute();
    glutPostRedisplay();  
}

void
setupMenu()
{
	char buf[100];
	int i, m, j, menu, control;
	int menus[nMenuSubshell];

	menu = 	glutCreateMenu(menuData);
	for (i = j = 0; i < nMenuSubshell; i++) {
		menus[i] = glutCreateMenu(menuData);
		for (m = -i; m <= i; m++, j++) {
			sprintf(buf, "m=%d", m);
			glutAddMenuEntry(strdup(buf), nMenuItems+j);
		}
	}

	control = glutCreateMenu(menuData);
	for (i = 0; i < nMenuItems; i++) {
		glutAddMenuEntry(menuItems[i], i);
	}

	glutSetMenu(menu);
	for (i = 0; i < nMenuSubshell; i++) {
		sprintf(buf, "l=%d", i);
		glutAddSubMenu(strdup(buf), menus[i]);
	}	
	glutAddSubMenu("Control", control);
	glutAttachMenu(GLUT_RIGHT_BUTTON);
}

int
main(int argc, char **argv)
{
  printf("SHG " __DATE__ " " __TIME__ "\n(C) Adomas Paltanavicius 2005\nSee http://adomas.org/shg/ for more information.\n\n");
  printf("Gnomovision comes with ABSOLUTELY NO WARRANTY; for details\n"
         "see GNU General Public License.\n\n");

  if (argc > 1) {
	  argL = atoi(argv[1]);
  }
  if (argc > 2) {
	  argM = atoi(argv[2]);
  }
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(700, 700);
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutCreateWindow("SHG");
  compute();
  glutDisplayFunc(display);
  glutKeyboardFunc(key);
  glutIdleFunc(idle);
  init();
  zprInit();
  setupMenu();
  glutMainLoop();
  return 0;
}
