/*
SHG -- Spherical Harmonics Generator.
Copyright (C) 2005, 2006 Adomas Paltanavicius

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*

there are problems with m=+-1 and oddness calculation. I don't
think that's bug in rendering code; the matter is that I am
not sure how to check for negative values. -- admp

*/

extern double legendre_Plm (const int, const int, const double);

void
Yslice(int l, int m, double theta, int slice, double *data, char *odd) 
{
	double cplx_real, cplx_imag;
	double under_root;
	double val = 1.0;
	int diff, sum;
	int mneg = 0; // True if m was negative at first
	int i;

	if (m < 0) {
		m = -m;
		if (m % 2 == 0) {
			val = -val;
			mneg = 1;
		}
	}
	diff = l - m;
	sum = l + m;
	under_root = (2 * l + 1) / (4 * M_PI);
	for (i = 2; i < diff; i++)
	  under_root *= i;
	for (i = 2; i < sum; i++)
	  under_root /= i;
	val *= sqrt (under_root);
	if (m != 0) {
		val *= M_SQRT2;
	}

	/* Legendre_Plm doesn't depend on phi; that's cool. We can optimize a lot.
	 */
	val *= legendre_Plm (l, m, cos (theta));

	for (i = 0; i < slice; i++) {
		double phi = (double)i/(double)slice * M_PI * 2.0;
		double q = 1.0;
		if (m != 0) {
			q = mneg ? cos(m * phi) : sin(m * phi);
		}
		cplx_real = q * val * cos (m * phi);
		cplx_imag = q * val * sin (m * phi);
		data[i] = sqrt (cplx_real * cplx_real + cplx_imag * cplx_imag);
		odd[i] = q*val<.001;
	}
}


// (a+bi)(c+di) = ac+adi+bci-bd
// (a+bi)(a+bi) = aa+abi+abi-bb = aa+2abi-bb
